package jrf.hbset;

import gov.nasa.jpf.jvm.ClassInfo;
import jrf.listener.HBSearchListener;
import jrf.listener.JRFListener;
import jrf.search.HBSearch;
import jrf.hbset.util.Key;
import jrf.hbset.util.MyMap;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;
import java.util.Iterator;
import jrf.hbset.util.*;


public class ListenerHBmanagement {
	// this is meaningless now... HASHMAP is out of date...
	//private static final String HBCHOICE="BITMAP";
	//private static final String HBCHOICE="HASHMAP";
	
	private static HBSet hb; // current HBset to store hb informations
	private static Stack<Node> stack; // current stack to store hb changes 
	//private static MyMap<Integer,ArrayList<State>> store; // used to store/restore HBset related information
												 // this used a lot of memory space, so store/restore is not good
												 // from the memory utilization point of view
	
	// defines HBrelated operation type to put into stack for undo
	private static enum OP { STATE, ADD, REMOVE, ALLOCATE_HB, ALLOCATE_HB_ENTRY, //GC, 
							// UNLOCK, 
							PRIOR_RELEASE, WRITTEN_BY, 
							REMOVE_STATIC_INIT, ADD_STARRTED_THREADS };
							
	private static boolean isSearchStarted=false;  // to indicate search is actually started by 1st advance
	
	private static MyMap<Key, StaticInfo> staticInitObj; // table of objects created by static initializer
													  // true is for volatile, false is for non-volatile
	private static MySet<Key> startedThreads; // table of threads already started
												// we don't need to worry about terminated thread
	public static enum RESTORE_ORDER { FIFO, LIFO };
	private static RESTORE_ORDER restore_order = RESTORE_ORDER.LIFO; // default DFS

	static 
	{
//		if ( HBCHOICE == "BITMAP" )
			hb = new BitVectorHBSet();
//		else if ( HBCHOICE == "HASHMAP")
//			hb = new HashMapHBSet();
//		else
//			assert(false):"Invalid option of HBSet";
			
		stack = new Stack<Node>();
//		store = new MyMap<Integer,ArrayList<State>>();
		staticInitObj = new MyMap<Key, StaticInfo>();
		startedThreads = new MySet<Key>();
	}
	
	public static HBSet getHB()
	{
		return hb;
	}
	
	public static void setRestore_FIFO_Order()
	{
		restore_order = RESTORE_ORDER.FIFO; // set to BFS
	}
	
	public static void setRestore_LIFO_Order()
	{
		restore_order = RESTORE_ORDER.LIFO; // set to BFS
	}
	
	
	public static boolean is_FIFO_Order()
	{
		return restore_order==RESTORE_ORDER.FIFO;
	}
	
	public static MySet<Key> getStaticObj()
    {
		MySet<Key> set = staticInitObj.keySet();
		MySet<Key> result = new MySet<Key>();
		for ( Key k : set )
            if (!staticInitObj.get(k).isDeleted) result.add(k);
		return result;
    }
	
	private static MyMap<Key, StaticInfo> copyStaticInitObj()
	{
		MyMap<Key, StaticInfo> staticCopy = new MyMap<Key, StaticInfo>(staticInitObj.size());
		MySet<Key> keySet = staticInitObj.keySet();
		for ( Key k : keySet )
		{
			StaticInfo sinfo = staticInitObj.get(k);
			staticCopy.put(k, new StaticInfo(sinfo.isVolatile, sinfo.isDeleted));
		}
		return staticCopy;
	}
	
	// static initialized object information
	static class StaticInfo {
		final boolean isVolatile;
		boolean isDeleted;
		
		StaticInfo(boolean isVolatile, boolean isDeleted)
		{
			this.isVolatile = isVolatile;
			this.isDeleted = isDeleted;
		}
		
		private StaticInfo markDeleted()
		{
			isDeleted = true;
			return this;
		}
		
		private StaticInfo unmarkDeleted()
		{
			isDeleted = false;
			return this;
		}
	}
	
	// HB information to store when state stored
	
	static class HBData {
		final HBSet hb;
		final Stack<Node> stack;
		final MySet<Key> startedThreads;
		final MyMap<Key, StaticInfo> staticInitObj;
		
		HBData(HBSet hb, Stack<Node> stack, MySet<Key> startThreads, MyMap<Key, StaticInfo> staticInit)
		{
			this.hb = hb;
			this.stack = stack;
			this.startedThreads = startThreads;
			this.staticInitObj = staticInit;
		}
	}
	
	
	// information to store in stack for furtner undo
	static class Node {
		final OP op;
		final Key set;
		final Key obj;
		
		Node(OP op, Key s, Key d)
		{
			this.op = op;
			this.set = s;
			this.obj = d;
		}
	}
	
	//
	//
	// PUBLICALLY ACCESSIBLE METHODS
	//
	//
	public static String getHBsize()
	{
		/*
		StringBuffer buf = new StringBuffer("Total HB entries,Total HB memory="
																+hb.get_table_entry()[0]+"*"
																+hb.get_table_entry()[1]+","
																+hb.get_table_size()+"\n");
																*/
		StringBuffer buf = new StringBuffer(hb.get_table_entry()[0]+"*"+hb.get_table_entry()[1]);
		//buf.append("Total HB memory="+hb.get_table_size()+"\n");
		//buf.append("Max Stored entry, Max Stored memory="+storedEntryMax+","+storedMemSize+"\n");
		return buf.toString();
	}
	
	
	public static void STATE_ADVANCE(int stateNum)//, HBSearch search)
	{
		if ( !isSearchStarted ) isSearchStarted = true;
		
		Log.log("\t\tSTATE_ADVANCE");
		stack.push(new Node(OP.STATE, new Key("",stateNum), null));
		Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+
				","+stack.peek().set+","+stack.peek().obj+")");
		
		//search.saveStateHBandStaticObject(stateNum, getHB(), getStaticObj());
	}
	
	
//	private static int storedEntryMax=0;
//	private static int storedMemSize=0;
	public static HBData STATE_STORE()
	{
		if ( !isSearchStarted ) return null;
		//Log.log("\t\tSTATE_STORE "+stateNum);
	
		HBData old =  new HBData( hb.clone(), 
				   (Stack<Node>)(stack.clone()), 
				   (MySet<Key>) startedThreads.clone(), 
				   copyStaticInitObj());
			
		return old;
		//ArrayList<State> list = store.get(stateNum);
		//if ( list == null ) list = new ArrayList<State>();
		
		// in case of DFS
		//list.add(newStore);
		
		// in case of BFS
		//list.add(0, newStore);
		
		//store.put(stateNum, list);
/*		
		if ( storedEntryMax < store.size() )
		{
			storedEntryMax = store.size();
			storedMemSize += (hb.get_table_size()+stack.size()*9);
		}
*/		
	}
	
	public static void STATE_RESTORE(Object oldData)
	{
		
		if ( !isSearchStarted ) return;
		//Log.log("\t\tSTATE_RESTORE "+stateNum);
		
		// restore to saved state
		//ArrayList<State> oldlist =  store.remove(stateNum);
		//State old = oldlist==null?null:oldlist.remove(oldlist.size()-1);
		//assert(old != null):"Restore to "+stateNum+" failed ";
		//if ( oldlist.size() > 0 ) store.put(stateNum, oldlist);
		HBData stored = (HBData)oldData;
		
		
		hb = stored.hb;
		stack = stored.stack;

		startedThreads = stored.startedThreads;
		MyMap<Key, StaticInfo> staticCopy = copyStaticInitObj();
		MySet<Key> existing = stored.staticInitObj.keySet();
		for ( Key k : existing )
			staticCopy.remove(k);
		
		existing = staticCopy.keySet();
		for ( Key k : existing )
			stored.staticInitObj.put(k, staticCopy.get(k));
		staticInitObj = stored.staticInitObj;
		
		// handle staticInitObj which is created after state store
		// we need to add all staticInitObj to started threads HBset
		// note. there is no need to store/restore startedThreads info 
		// since thread whose start is undone will not be scheduled 
		// after restore until started again, and at the new start, 
		// all HB information will be reinitialized 
		MySet<Key> staticObjs = staticInitObj.keySet();
		Iterator<Key> it = staticObjs.iterator();
		while ( it.hasNext() )
		{
			Key k = it.next();
			StaticInfo sinfo = staticInitObj.get(k);
			if ( !sinfo.isDeleted && sinfo.isVolatile && !hb.is_hb_member(k) )
			{			
				hb.allocate(k, true);
			}
			else if ( !sinfo.isDeleted && !sinfo.isVolatile )
			{					
				if ( !hb.is_column_member(k) ) 
					hb.allocate(k, false);
				Iterator<Key> itt = startedThreads.iterator();
				while ( itt.hasNext() )
				{
					Key ts = itt.next();
					hb.add(ts, k);
				}
			}
		}
	}
	
	
	public static void STATE_BACKTRACK(int stateNum)
	{
		Log.log("\t\tSTATE_BACKTRACK");
		
		assert ( isSearchStarted ):"STATE_BACKTRACK without prior STATE_ADVANCE";
		Log.logStack("\t\tSTART <STATE_BACKTRACK>");
		//MyMap<Key, Key> classObj = new MyMap<Key, Key>();
		
		// undo until state reached...
		for ( Node t=stack.pop(); true ; t=stack.pop() )
		{
			// logging purpose....
			 if ( t.op != OP.ADD && t.op != OP.REMOVE )
				 Log.logStack("\t\t\tSTACK_POP("+t.op+","+t.set+","+t.obj+")");
			if ( t.op == OP.STATE )
			{
				// state is reached
				if ( t.set.equals(new Key("",stateNum)) ) 
				{
					// save STATE mark for further state backtracking
					Log.logStack("\t\t\tSTACK_PUSH_AGAIN("+t.op+","+t.set+","+t.obj+")");
					stack.push(t);
					break;
				}
				else
				{
					Log.logStack("\t\t\tONE_MORE_BACKTRACK since "+t.set+" != "+stateNum);
					//continue;
				}
			}
			// undo allocate_hb
			else if ( t.op == OP.ALLOCATE_HB )
			{
				/* don't need this any more since we don't undo staticInit
				if ( t.obj != null ) // it's class 
				{
					Log.logStack("t\t\t(NEEDS REDONE,"+t.set+","+t.obj+")");
					classObj.put(t.set, t.obj); // save for REDO
				}
				*/
				Log.logStack("\t\t\t(UNDO_ALLOCATE_HB,"+t.set+","+t.obj+")");
				hb.deallocate(t.set, true);
			}
			// undo allocate_hb_entry
			else if ( t.op == OP.ALLOCATE_HB_ENTRY )
			{
				/* don't need this any more since we don't undo staticInit
				if ( t.obj != null ) // it's class 
				{
					Log.logStack("t\t\t(NEEDS REDONE,"+t.set+","+t.obj+")");
					classObj.put(t.set, t.obj); // save for REDO
				}
				*/ 
				Log.logStack("\t\t\t(UNDO_ALLOCATE_HB_ENTRY,"+t.set+","+t.obj+")");
				hb.deallocate(t.set, false);
			}
			// undo remove
			else if ( t.op == OP.REMOVE )
			{
				Log.logStack("\t\t\t(UNDO_REMOVE,"+t.set+","+t.obj+")");
				hb.add(t.set, t.obj);
			}
			// undo add
			else if(  t.op == OP.ADD )
			{
				Log.logStack("\t\t\t(UNDO_ADD,"+t.set+","+t.obj+")");
				hb.remove(t.set, t.obj);
			}
			// undo prior_release set
			else if ( t.op == OP.PRIOR_RELEASE )
			{
				Log.logStack("\t\t\t(RESTORE_PRIOR_RELEASE,"+t.set+","+t.obj+")");
				hb.set_prior_release(t.obj, t.set);
			}
			// undo written_by set
			else if ( t.op == OP.WRITTEN_BY )
			{
				Log.logStack("\t\t\t(RESTORE_WRITTEN_BY,"+t.set+","+t.obj+")");
				hb.set_written_by(t.obj, t.set);
			}
			// undo remove from static_init since the write to statoc init obj will be undoen
			else if ( t.op == OP.REMOVE_STATIC_INIT )
			{
				Log.logStack("\t\t\t(RESTORE_REMOVE_STATIC_INIT,"+t.set+","+t.obj+")");
				StaticInfo old = staticInitObj.get(t.set);
				staticInitObj.put(t.set, old.unmarkDeleted());
			}
			// undo thread start
			else if ( t.op == OP.ADD_STARRTED_THREADS )
			{
				Log.logStack("\t\t\t(RESTORE_ADD_STARRTED_THREADS,"+t.set+","+t.obj+")");
				startedThreads.remove(t.set);
			}
		}
		Log.logStack("\t\tDONE <STATE_BACKTRACK>");
	}
	
	/*
	 * threadStart(t)
	 *     is called when new thread starts...copy parent thread's HB set into children's HB set 
	 */
	public static  void THREAD_START(String parentT,  String childT, boolean staticInit)
	{
	    Log.log("\t\t"+"[THREAD_START]\""+parentT+"\" STARTS \""+childT+"\"");	
	    
	    //boolean staticInit = AssertionalReasoningListener.isWithinStaticInit();
	    Key parent = new Key(parentT, 0);
		Key child = new Key(childT, 0);
			
		if ( isSearchStarted ) 
		{
			if ( childT.compareTo(parentT) == 0 ) // start of main thread
			{
				List<KeyPair> existing = hb.allocate(parent, true);
				if ( !staticInit ) // we don't need to undo hb operations inside staticInit 
				{
					if ( existing == null )
					{
						stack.push(new Node(OP.ALLOCATE_HB, parent, null));
						Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+
								","+stack.peek().set+":"+stack.peek().set+","+stack.peek().obj+")");
					}
					else {
						Iterator<KeyPair> pair = existing.iterator();
						while  ( pair.hasNext() )
						{
							KeyPair p = pair.next();
							stack.push(new Node(OP.REMOVE, p.obj, p.set));
							Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","
								+stack.peek().set+":"+stack.peek().set+","+stack.peek().obj+")");
						}
					}
				}
				
				existing = hb.allocate(parent, false);
				if ( !staticInit ) // we don't need to undo hb operations inside staticInit 
				{
					if ( existing == null )
					{
						stack.push(new Node(OP.ALLOCATE_HB_ENTRY, parent, null));
						Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+
								","+stack.peek().set+":"+stack.peek().set+","+stack.peek().obj+")");
					}
					else {
						Iterator<KeyPair> pair = existing.iterator();
						while  ( pair.hasNext() )
						{
							KeyPair p = pair.next();
							stack.push(new Node(OP.REMOVE, p.obj, p.set));
							Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","
								+stack.peek().set+":"+stack.peek().set+","+stack.peek().obj+")");
						}
					}
				}
			}
			else
			{
				// allocate hb for child if necessary
				if ( !hb.is_hb_member(child) )
					hb.allocate(child, true); // the returned will be always null since we already checked "is_hb_member"
				
				MySet<Key> b1=null, b2=null;
				if ( !staticInit )
					b1 = hb.get_hb_set(child); // b1 = hb(child)
				
				hb.acquire(child, parent);
				
				if ( !staticInit )
				{
					b2 = hb.get_hb_set(child);
					Iterator<Key> it = b1.iterator();
					while ( it.hasNext() )
					{
						Key k = it.next();
						b2.remove(k);
					}
					
					it = b2.iterator();
					while ( it.hasNext() )
					{
						Key k = it.next();
						if ( !child.equals(k) )
						{
							stack.push(new Node(OP.ADD, child, k));
							Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","+
									stack.peek().set+","+stack.peek().obj+")");
							// save acquiring history
							// do we need this??
							//ACHSet.add(k, child, parent, ACHSet.ACCESS_TYPE.T_START);
						}
					}
				}
			}
		}
		else
		{
			if ( childT.compareTo(parentT) == 0 ) // start of main thread
			{
				hb.allocate(parent, true);
				hb.allocate(parent, false);
			}
			else
			{
				// allocate hb for child if necessary
				if ( !hb.is_hb_member(child) )
					hb.allocate(child, true); // the returned will be always null since we already checked "is_hb_member"
				
				// copy hb(parent) to hb(child)
				hb.acquire(child, parent);
			}
		}
		
		MySet<Key> keySet = staticInitObj.keySet();
		Iterator<Key> it = keySet.iterator();
	
		while ( it.hasNext() )
		{
			Key obj = it.next();
			if ( !staticInitObj.get(obj).isVolatile ) // non-volatile
			{
				hb.add(child, obj);
				if ( !staticInit )
				{
					stack.push(new Node(OP.ADD, child, obj));
					Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","+
						stack.peek().set+","+stack.peek().obj+")");
				}
				// save acquiring history
				// Do we need this???
				// ACHSet.add(obj, child, null, ACHSet.ACCESS_TYPE.STATIC);
			}
		}
		startedThreads.add(child);
		if ( !staticInit ) // why this is needed?
		{
			stack.push(new Node(OP.ADD_STARRTED_THREADS, child, null));
			Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","+
					stack.peek().set+","+stack.peek().obj+")");
		}
		
		// add to future access key
		HBSearchListener.addFutureAccessKey(parent, true);
		HBSearchListener.addFutureAccessKey(child, true);
	}

	/*
	 * newObj(thread->obj)
	 *     instantiate HB set for srcO->srcF 
	 */
	public static void NEW_OBJECT(	String thread, 
									String obj, 
									int fn, 
									boolean isVolatile, 
									boolean staticInit)//, 
									//boolean isFinal, 
									//boolean isAccessible)	
	{	
		Log.log("\t\t[NEW_OBJECT]"+obj+"->"+fn+" by "+thread+", isVolatile="+isVolatile+", isStaticInit="+staticInit);	
		Key t = new Key(thread, 0);
		Key o = new Key(obj, fn);
		
		if ( isSearchStarted && !staticInit ) 	// Class obj should not be deallocated
									// since it will not be reloaded
		{
			List<KeyPair> existing = hb.allocate(o, isVolatile);
			
			if ( existing != null )
			{
				Iterator<KeyPair> pair = existing.iterator();
				while  ( pair.hasNext() )
				{
					KeyPair p = pair.next();
					stack.push(new Node(OP.REMOVE, p.obj, p.set));
					Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","
							+stack.peek().set+":"+stack.peek().set+","+stack.peek().obj+")");
				}
			}
			
			stack.push(new Node(isVolatile?OP.ALLOCATE_HB:OP.ALLOCATE_HB_ENTRY, o, null));
			Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","
					+stack.peek().set+","+stack.peek().obj+")");
				
			if ( !isVolatile ) 
			{
				hb.add(t, o);
				// save history of change
				stack.push(new Node(OP.ADD, t, o));
				Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","
						+stack.peek().set+","+stack.peek().obj+")");
				// save acquiring history
				// do we need this??
				// ACHSet.add(o, t, null, ACHSet.ACCESS_TYPE.INIT);
			}
		}
		else
		{
			hb.allocate(o, isVolatile);
			if ( !isVolatile ) hb.add(t, o);
		}
		
		if ( staticInit ) 
		{ 		
			staticInitObj.put(o, new StaticInfo(isVolatile, false));
			Log.log("\t\t [NEW_OBJECT] add "+o+" to staticInitObj[] set" +"["+staticInitObj.containsKey(o)+"]");
			
			if ( !isVolatile )
			{
				Iterator<Key> it = startedThreads.iterator();
				while ( it.hasNext() )
				{
					Key ts = it.next();
					hb.add(ts, o);
					Log.log("\t\t [NEW_OBJECT] add "+o+" to already started thread "+ts +"["+staticInitObj.containsKey(o)+"]");
				}
			}
		}
		
		// save source position of allocation
		//if ( o.getLocation() == null )
		//	o.setLocation(HBDataRaceException.getSourceLocation());
	}
	
	/*
	 * READ
	 * 	called before read operation & lock acquire(with isVolatile set)
	 */
	public  static void READ(String thread, String obj, int fn, boolean isVolatile, boolean isStaticInit)  
	{

		Key t = new Key(thread, 0);
		Key o = new Key(obj, fn);
		// add to future access key
		HBSearchListener.addFutureAccessKey(t, true);
		HBSearchListener.addFutureAccessKey(o, false);
try{
		boolean isClass = obj.startsWith("java.lang.Class@") || (obj.indexOf("@")==-1 && ClassInfo.getResolvedClassInfo(obj) != null) ; 
		Log.log("\t\t[READ]-> "+obj+"->"+fn+" by "+thread+"(isVolatile="+isVolatile+", isClass="+isClass+", isStaticInit="+isStaticInit+")");
		
		if ( isVolatile )
		{
			if ( isSearchStarted && !isStaticInit )
			{
				// save history of change
				MySet<Key> b1, b2;
				b1 = hb.get_hb_set(t);
				hb.acquire(t, o);
				b2 = hb.get_hb_set(t);
				Iterator<Key> it = b1.iterator();
				while ( it.hasNext() )
				{
					Key k = it.next();
					b2.remove(k);
				}
				
				it = b2.iterator();
				while ( it.hasNext() )
				{
					Key k = it.next();
					if ( !k.equals(t) )
					{
						stack.push(new Node(OP.ADD, t, k));
						Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","
								+stack.peek().set+","+stack.peek().obj+")");
						// save acquiring history
						/* This is to be used only when actual addition to hb set */
						ACHSet.add(k, t, o);
						/**/
					}
				}
			}
			else
				hb.acquire(t, o);
		}
		else
			if ( !isClass || fn != 0 ) 
			{
				Log.log("\t\t\t NORACE("+t+","+o+")");
				hb.norace(t, o); 			
			}
}
catch(HBDataRaceException e)
{
	JRFListener.raceHandler(e);
}
	}

	/*
	 * WRITE
	 * 	called at write operation & lock release(with isVolatile flag set)
	 */
	public  static void WRITE(String thread, String obj, int fn, boolean isVolatile, boolean isStaticInit) 
	{		
		Key t = new Key(thread, 0);
		Key o = new Key(obj, fn);
		
		// add to future access key
		HBSearchListener.addFutureAccessKey(t, true);
		HBSearchListener.addFutureAccessKey(o, false);
try{
		Log.log("\t\t[WRITE]-> "+obj+"->"+fn+" by "+thread+"["+staticInitObj.containsKey(o)+","
				+isStaticInit+"]");
		if ( isVolatile )
		{
			Key old = hb.set_prior_release(t, o);
			if ( isSearchStarted && !isStaticInit )
			{
				// save history of change
				MySet<Key> b1, b2;
				b1 = hb.get_hb_set(o);

				hb.release(t, o);
				b2 = hb.get_hb_set(o);

				Iterator<Key> it = b1.iterator();
				while ( it.hasNext() )
				{
					Key k = it.next();
					b2.remove(k);
				}
				
				it = b2.iterator();

				while ( it.hasNext() )
				{
					Key k = it.next();					
					if ( !k.equals(o) )
					{
						stack.push(new Node(OP.ADD, o, k));
						Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","
								+stack.peek().set+","+stack.peek().obj+")");
					}
				}
				
				if ( old == null || !old.equals(t) )
				{
					stack.push(new Node(OP.PRIOR_RELEASE, o, old));
					Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","
							+stack.peek().set+","+stack.peek().obj+")");
				}
			}
			else
				hb.release(t, o);
			
		}
		else if ( !staticInitObj.containsKey(o) || staticInitObj.get(o).isDeleted  || !isStaticInit )
		{
			Log.log("\t\t\t NORACE("+t+","+o+")");
			hb.norace(t, o); 
			
			Key old = hb.set_written_by(t, o);
			if ( isSearchStarted && !isStaticInit )
			{
				// save history of change
				MySet<Key> a1, a2;
				a1 = hb.get_key_set(o);
				hb.invalidate(t, o);
				a2 = hb.get_key_set(o);
				
				Iterator<Key> it = a2.iterator();
				while ( it.hasNext() )
				{
					Key k = it.next();
					a1.remove(k);
				}
				
				it = a1.iterator();
				while ( it.hasNext() )
				{
					stack.push(new Node(OP.REMOVE, it.next(), o));
					Log.logStack("\t\t\tSTACK_PUSH_SKIPPED("+stack.peek().op+","
							+stack.peek().set+","+stack.peek().obj+")");
				}
				if ( old == null || !old.equals(t) )
				{
					stack.push(new Node(OP.WRITTEN_BY, o, old));
					Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","
							+stack.peek().set+","+stack.peek().obj+")");
				}
			}
			else
				hb.invalidate(t, o);
		}	
	//	else
	//		System.out.println("\t\t[SKIP]-> "+obj+"->"+fn+" by "+thread+"["+staticInitObj.containsKey(o)+","+isStaticInit+"]");
		
		if ( staticInitObj.containsKey(o) && !isStaticInit ) 
		{
			StaticInfo old = staticInitObj.get(o);
			staticInitObj.put(o, old.markDeleted());
			
			stack.push(new Node(OP.REMOVE_STATIC_INIT, o, null));
			Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","+
					stack.peek().set+","+stack.peek().obj+")");
		}
}
catch(HBDataRaceException e)
{
	JRFListener.raceHandler(e);
}
	}
	
	/*
	 * LOCK
	 * 	called before read operation
	 */
	public  static void LOCK(String thread, String obj, boolean isStaticInit)
	{	
		Log.log("\t\t[LOCK]-> "+obj+" by "+thread);
		
		Key o = new Key(obj, 0);
		if ( !hb.is_hb_member(o) ) // allocate hb_row_entry if it's first locking
		{
			hb.allocate(o, true);
			if ( !isStaticInit )
			{
				stack.push(new Node(OP.ALLOCATE_HB, o, null));
				Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","
					+stack.peek().set+","+stack.peek().obj+")");
			}
		}
		READ(thread, obj, 0, true, isStaticInit);
	}
	
	/*
	 * UNLOCK
	 * 	called at write operation
	 */
	public  static void UNLOCK(String thread, String obj, boolean isStaticInit) 
	{	
		Log.log("\t\t[UNLOCK]-> "+obj+" by "+thread);
		
		Key o = new Key(obj, 0);
		if ( !hb.is_hb_member(o) ) // allocate hb_row_entry if it's first locking
		{
			hb.allocate(o, true);
			if ( !isStaticInit )
			{
				stack.push(new Node(OP.ALLOCATE_HB, o, null));
				Log.logStack("\t\t\tSTACK_PUSH("+stack.peek().op+","
					+stack.peek().set+","+stack.peek().obj+")");
			}
		}
		WRITE(thread, obj, 0, true, isStaticInit);//, false);
	}
	
	/*
	 * WRITTEN_BY
	 * return true if obj.fn is recently written by thread 
	 */
	public static boolean WRITTEN_BY(String thread, String obj, int fn)
	{
		Key t = new Key(thread, 0);
		Key o = new Key(obj, fn);
		Key old = hb.get_written_by(o);
		if ( old != null && old.equals(t) ) return true;
		else return false;
	}
	
	/*
	 * PRIOR_RELEASE
	 * return true if obj.fn is recently written by thread 
	 */
	public static boolean PRIOR_RELEASE(String obj, int fn)
	{
		Key o = new Key(obj, fn);
		Key old = hb.get_prior_release(o);
		if ( old != null ) return true;
		else return false;
	}
	
	public static MySet<Key> CANDIDATE_AGENTS(Key o)
	{
		return hb.get_key_set(o);
	}
	
	public static boolean HAS_REPRESENTATION(Key e)
	{
		if ( hb.is_column_member(e) && hb.get_key_set(e).size() > 1 ) return true;
		else return false;
		//return hb.is_column_member(a)&&(!hb.is_column_member(e)||hb.get_key_set(e).size()==1);
	}
	
	public static boolean IS_SYNCHADDR_OR_THREAD(Key o)
	{
		return hb.is_hb_member(o);
	}
	
	public static boolean IS_ADDR(Key o)
	{
		return hb.is_column_member(o);
	}
	
	public static MySet<Key> ADDR()
	{
		return hb.get_column();
	}
	
	public static MySet<Key> SYNCHADDR()
	{
		return hb.get_row();
	}
}
