package jrf.hbset.util;

import gov.nasa.jpf.jvm.ClassInfo;
import gov.nasa.jpf.jvm.ElementInfo;
import gov.nasa.jpf.jvm.FieldInfo;
import gov.nasa.jpf.jvm.ThreadInfo;

import jrf.hbset.ListenerHBmanagement;
import static jrf.hbset.util.Utility.*;

public class AbstractArrayImpl {
	public static final int LAZY=-1;
	/*
	 * WRITE stub for dealing with TrustedClass option
	 * 	called at write operation
	 */
	public static void WRITE(ThreadInfo ti, ElementInfo ei, FieldInfo fi, int fn, boolean isVolatile, boolean isStaticInit) 
	{
/*		
System.out.println("[WRITE] "+ei.toString()+"."+fn+","+isVolatile+","+isStaticInit+"("+
				", isImmutable="+isImmutable(!(ei.toString().contains("@")), fi, fn)+
				", isInteresting="+Configuration.isInteresting(ei.toString()) +
				", isAccessible="+isAccessible(fi) +
				", isExcluded="+isExcluded(ei.toString(), fn, fi));	
	*/			
		if ( isVolatile || 
				( !isImmutable(!(ei.toString().contains("@")), fi, fn) &&  // "Class" Object is immutable
				  ( Configuration.isInteresting(ei.toString()) || isAccessible(fi) ) &&
				  !isExcluded(ei.toString(), fn, fi)
				) 
			)			
		{			
			// dealing with ArrayAbstraction
			// from here
			Key array = new Key(ei.toString(), 0);
			Key element = new Key(ei.toString(), fn);
			if ( !isVolatile  && ei.isArray() && fn>0 && !ListenerHBmanagement.HAS_REPRESENTATION(element)) // 
			{			
				Key thread = new Key(getThreadObject(ti), 0);
				MySet<Key> arrayAgent = ListenerHBmanagement.CANDIDATE_AGENTS(array);
				if ( arrayAgent.size() != 2 /* self and current thread */ || !arrayAgent.contains(thread) ) // need to allocate hb_column for this array element
				{
					NEW_OBJECT(ti, ei, fi, fn, isVolatile, isStaticInit);
				}
				else // we do write to array instead of individual element
				{
					ListenerHBmanagement.WRITE(	getThreadObject(ti), 
							ei.toString(), 
							LAZY, 
							isVolatile, 
							isStaticInit);
				}
			}
			else
			// to here 
			{				
				ListenerHBmanagement.WRITE(	getThreadObject(ti), 
										ei.toString(), 
										fn, 
										isVolatile, 
										isStaticInit);
			}
		}
	}
	
	/*
	 * READ stub for dealing with TrustedClass option
	 * 	called at read operation
	 */
	public static void READ(ThreadInfo ti, ElementInfo ei, FieldInfo fi, int fn, boolean isVolatile, boolean isStaticInit) 
	{
/*		
System.out.println("[READ] "+ei.toString()+"."+fn+","+isVolatile+","+isStaticInit+"("+
		", isImmutable="+isImmutable(!(ei.toString().contains("@")), fi, fn)+
		", isInteresting="+Configuration.isInteresting(ei.toString()) +
		", isAccessible="+isAccessible(fi) +
		", isExcluded="+isExcluded(ei.toString(), fn, fi));		
*/
		if ( isVolatile || 
				( !isImmutable(!(ei.toString().contains("@")), fi, fn) && 
				  ( Configuration.isInteresting(ei.toString()) || isAccessible(fi) ) &&
				  !isExcluded(ei.toString(), fn, fi)
				) 
			)	
		{	
			// dealing with ArrayAbstraction
			// from here
			//Key array = new Key(ei.toString(), 0);
			Key element = new Key(ei.toString(), fn);
			if ( !isVolatile && ei.isArray() && fn>0 && !ListenerHBmanagement.HAS_REPRESENTATION(element)) // 
			{
				
				 // we do read from array instead of individual element
				ListenerHBmanagement.READ(	getThreadObject(ti), 
							ei.toString(), 
							LAZY, 
							isVolatile, 
							isStaticInit);
			}
			else
			// to here 
				
				ListenerHBmanagement.READ(	getThreadObject(ti), 
										ei.toString(), 
										fn, 
										isVolatile, 
										isStaticInit);
		}		
	}
	
	public static void NEW_OBJECT(ThreadInfo ti, ElementInfo ei, FieldInfo fi, int fn, boolean isVolatile, boolean staticInit)	
	{
		/*
System.out.println("[NEW_OBJECT] "+ei.toString()+"."+fn+","+isVolatile+","+staticInit+"("+
				", isImmutable="+isImmutable(false, fi, fn)+
				", isInteresting="+Configuration.isInteresting(ei.toString()) +
				", isAccessible="+isAccessible(fi) +
				", isExcluded="+isExcluded(ei.toString(), fn, fi));		
			*/		
		if ( isVolatile || 
				( !isImmutable(false, fi, fn) && 
				  ( Configuration.isInteresting(ei.toString()) || isAccessible(fi) ) &&
				  !isExcluded(ei.toString(), fn, fi)
				) 
			)	
		{
			ListenerHBmanagement.NEW_OBJECT(	getThreadObject(ti), 
												ei.toString(), 
												fn, 
												isVolatile, 
												staticInit);
		}
	}
	
	public static void NEW_CLASS(ThreadInfo ti, ClassInfo ci, FieldInfo fi, int fn, boolean isVolatile, boolean staticInit)	
	{
		/*
System.out.println("[NEW_CLASS] "+ci.getName()+"."+fn+","+isVolatile+","+staticInit+"("+
				", isImmutable="+isImmutable(true, fi, fn)+
				", isInteresting="+Configuration.isInteresting(ci.getName()) +
				", isAccessible="+isAccessible(fi) +
				", isExcluded="+isExcluded(ci.getName(), fn, fi));	
				*/
		if ( isVolatile || 
				( !isImmutable(true, fi, fn) && 
				  ( Configuration.isInteresting(ci.getName()) || isAccessible(fi) ) &&
				  !isExcluded(ci.getName(), fn, fi)
				) 
			)		
		{
			ListenerHBmanagement.NEW_OBJECT(	getThreadObject(ti), 
												ci.getName(), 
												fn, 
												isVolatile, 
												staticInit);
		}
	}
}
