package jrf.hbset.util;

import jrf.hbset.ListenerHBmanagement;
import jrf.hbset.util.Utility.Comparison;

//stores future access for each states
public class FutureAccess {

	MySet<Key> current;
	MySet<FutureAccess> children;
	MySet<MyMap<Key, MySet<Key>>> h_future;
	boolean isSubtreeProcessed;
	int isVisited;
	int id;
	
	public FutureAccess(int state_number)
	{
		current = new MySet<Key>();
		children = new MySet<FutureAccess>();
		h_future = new MySet<MyMap<Key, MySet<Key>>>();
		isSubtreeProcessed = false;
		isVisited = 0;
		id = state_number;
	}

	public void setProcessed()
	{
		isSubtreeProcessed = true;
	}
	
	private boolean isSubtreeProcessed(MySet<FutureAccess> exclusion)
	{
		if ( isSubtreeProcessed ) return true;
		if ( !isVisited() ) 
			return false;
		
		MySet<FutureAccess> addedChildren = new MySet<FutureAccess>(exclusion);
		addedChildren.addAll(children);
		for( FutureAccess fa : children )
		{
			if ( !exclusion.contains(fa) && !fa.isSubtreeProcessed(addedChildren) )
				return false;
		}
		setProcessed();
		return true;
	}
	
	public boolean isSubtreeProcessed()
	{
		MySet<FutureAccess> me = new MySet<FutureAccess>();
		me.add(this);
		return isSubtreeProcessed(me);
	}
	
	public void setVisited()
	{
		++isVisited;
	}
	
	public boolean isVisited()
	{
		return isVisited>1;
	}
	
	public void addAccess(Key k)
	{
		current.add(k);
	}
	
	public void addChild(FutureAccess ch)
	{
		children.add(ch);
	}
	
	public void addHfuture(MyMap<Key, MySet<Key>> hFuture)
	{
		//h_future.clear();
		h_future.add(hFuture);
	}
	
	public void changeHfuture(MyMap<Key, MySet<Key>> oldH, MyMap<Key, MySet<Key>> newH)
	{
		
		if ( h_future.remove(oldH) ) //&& !isCovered(newH) )
			h_future.add(newH);
		/*	
		h_future.clear();
		h_future.add(newH);
		*/
	}
	
	public boolean isCovered(MyMap<Key, MySet<Key>> hFuture)
	{
		MySet<Key> future_keys = hFuture.keySet();
		MySet<Key> future_set = getFutureAccess(null);
		
		for ( MyMap<Key, MySet<Key>> stored_h_future : h_future )
		{
			boolean isCovered = true;
			boolean smaller = false;
			for ( Key future : future_keys )
			{
				MySet<Key> all = stored_h_future.get(future);
				if ( all == null ) continue;
				MySet<Key> stored = new MySet<Key>(all);
				for ( Key k : all )
					if ( !future_set.contains(k) ) 
					{
						stored.remove(k);
						smaller = true;
					}

				MySet<Key> current = hFuture.get(future);

				Comparison result =  jrf.hbset.util.Utility.compare(current, stored); 
				if ( result != Comparison.EQUAL && result != Comparison.BIGGER )
				{
					Log.logSearch("\t\t[NOTICE]  DIFF because of key \""+future+"\".", null);
					Log.logSearch("\t\t\t\tSTORED="+stored, null);
					Log.logSearch("\t\t\t\tCURRENT="+current, null);
					
					isCovered = false;
					break;
				}
			}
			if ( isCovered )
			{	
				if ( smaller )
					changeHfuture(stored_h_future, hFuture);
				return true;	
			}
		}
		return false;
	}
	
	public MySet<Key> getFutureAccess(MySet<FutureAccess> exclusion)
	{
		if ( !isSubtreeProcessed() ) 
		{
			MySet<Key> all = ListenerHBmanagement.ADDR();
			all.addAll(ListenerHBmanagement.SYNCHADDR());
			return all;
		}
		else
		{
			if ( exclusion == null )
				exclusion = new MySet<FutureAccess>();
			// to prevent infinite recursion
			MySet<FutureAccess> additional = new MySet<FutureAccess>();
			exclusion.add(this);
			additional.addAll(exclusion);
			additional.addAll(children);
			
			for ( FutureAccess ch : children )
			{
				if ( !exclusion.contains(ch) )
					current.addAll(ch.getFutureAccess(additional));
			}
			// TO DO debug!!!
			children.clear(); //only calculate once
			return current;
		}
	}
	
	public MySet<MyMap<Key, MySet<Key>>> getHfuture()
	{
		return h_future;
	}
}