package jrf.hbset.util;

import gov.nasa.jpf.jvm.ClassInfo;
import gov.nasa.jpf.jvm.FieldInfo;

public class Key  {
	public final int obj;
	public final int field;
	
	static final MyMap<String, Integer> name = new MyMap<String, Integer>();
	static final MyMap<Integer, String> id = new MyMap<Integer, String>();
	static final MyMap<Integer, String> location = new MyMap<Integer, String>();
	
	static private int index=0;
	static final int SIZE_OF_KEY = 4;
	
	public Key(String objStr, int field)
	{
		Integer val = name.get(objStr);
		if ( val == null )
		{
			val = Integer.valueOf(index++);
			name.put(objStr, val);
			id.put(val, objStr);
		}
		
		this.obj = val.intValue();
		this.field = field;
	}
	
	public Key(Key o)
	{
		this.obj = o.obj;
		this.field = o.field;
	}
	
	public Key(Key o, int field)
	{
		this.obj = o.obj;
		this.field = field;
	}
	
	@Override
	public boolean equals(Object o)
	{
		if ( o == null ) return false;
		if ( !(o instanceof Key) ) return false;
		return obj == ((Key)o).obj && field==((Key)o).field;
	}
	
	@Override
	public int hashCode()
	{
		return obj<<16|field;
	}

	
	public void setLocation(String loc)
	{	
		Integer val = new Integer(obj);
		if ( id.get(val) == null )
			assert (false):"Key "+this+" is not in table yet.";
		location.put(val, loc);	
	}
	
	public String getLocation()
	{
		Integer val = new Integer(obj);
		if ( id.get(val) == null )
			assert (false):"Key "+this+" is not in table yet.";
		return location.get(val);	
	}
	
	
	public String toNameStringWithLocation()
	{
		String idStr = id.get(new Integer(obj));
		String[] cn = idStr.split("@");
		ClassInfo ci = ClassInfo.getResolvedClassInfo(cn[0]);
		
		if ( cn.length == 1 )  // static fields
		{
			FieldInfo fi = ci.getStaticField(field-1);
			return id.get(new Integer(obj))+(field==0?"":("."+fi.getName()))+" from "+getLocation();
		}
		else if ( idStr.charAt(0) == '[' ) // array
		{
			return id.get(new Integer(obj))+"["+toArrayIndex(field)+"]"+" from "+getLocation();
			//return id.get(new Integer(obj))+"["+((field-2)>0?(field-2):((field==AbstractArrayImpl.LAZY)?"LAZY":""))+"]"+" from "+getLocation();
		}
		else // instance field
		{
			FieldInfo fi = ci.getInstanceField(field-1);
			return id.get(new Integer(obj))+(field==0?"":("."+fi.getName()))+" from "+getLocation();
		}
	}
	
	private String toArrayIndex(int index)
	{
		if ( index==0 ) return "";
		else if ( index==AbstractArrayImpl.LAZY ) return "LAZY";
		else return ""+(index-1);
	}
	
	public String toNameString()
	{
		String idStr = id.get(new Integer(obj));
		String[] cn = idStr.split("@");
		ClassInfo ci = ClassInfo.getResolvedClassInfo(cn[0]);
		
		if ( cn.length == 1 )  // static fields
		{
			FieldInfo fi = ci.getStaticField(field-1);
			return id.get(new Integer(obj))+(field==0?"":("."+fi.getName()));
		}
		else if ( idStr.charAt(0) == '[' ) // array
		{
			return id.get(new Integer(obj))+"["+toArrayIndex(field)+"]";
		}
		else // instance field
		{
			FieldInfo fi = ci.getInstanceField(field-1);
			return id.get(new Integer(obj))+(field==0?"":("."+fi.getName()));
		}
	}

	public String toName()
	{
		String idStr = id.get(new Integer(obj));
		String[] cn = idStr.split("@");
		ClassInfo ci = ClassInfo.getResolvedClassInfo(cn[0]);
		
		if ( cn.length == 1 )  // static fields
		{
			FieldInfo fi = ci.getStaticField(field-1);
			return id.get(new Integer(obj))+(field==0?"":("."+fi.getName()));
		}
		else if ( idStr.charAt(0) == '[' ) // array
		{
			String loc = location.get(new Integer(obj));
			if ( loc.startsWith("\"") ) loc = loc.substring(1);
			
			int i = loc.indexOf("new");
			if ( i <= 0 ) 
			{
				i = loc.indexOf("{");
				if ( i <= 0 )
				{
					System.err.println("cannot locate \"new\" or { in "+loc);
					loc = id.get(new Integer(obj));
					return loc.trim()+"["+toArrayIndex(field)+"]";
				}
			}
			loc = loc.substring(0, i-1);
			
			i = loc.indexOf("=");
			if ( i <= 0 )
			{
				System.err.println("cannot locate \"=\" or { in "+loc);
				loc = id.get(new Integer(obj));
				return loc.trim()+"["+toArrayIndex(field)+"]";
			
			}
			loc = loc.substring(0, i-1);
			
			i = loc.lastIndexOf(" ");
			if ( i == loc.length()-1 ) loc = loc.substring(0, i);
			i = loc.lastIndexOf(" ");
			if ( i >= 0 ) loc = loc.substring(i+1, loc.length());

			return loc.trim()+"["+toArrayIndex(field)+"]";
		}
		else // instance field
		{
			FieldInfo fi = ci.getInstanceField(field-1);
			String name = id.get(new Integer(obj));
			return name.substring(0, name.lastIndexOf("@")+1)+(field==0?"":("."+fi.getName()));
		}
	}
	
	@Override
	public String toString()
	{
		return id.get(new Integer(obj))+(field==0?"":(field==AbstractArrayImpl.LAZY?".LAZY":("."+field)));
		//return obj+"."+field;
	}
	
	public boolean isField()
	{
		return field!=0;
	}
}
