package jrf.listener;

import java.util.ArrayList;

import jrf.hbset.ListenerHBmanagement;
import jrf.hbset.util.FutureAccess;
import jrf.hbset.util.Key;
import jrf.hbset.util.Log;
import jrf.hbset.util.MyMap;
import jrf.hbset.util.Utility.Comparison;
import jrf.hbset.util.MySet;


import gov.nasa.jpf.PropertyListenerAdapter;
import gov.nasa.jpf.jvm.JVM;
import gov.nasa.jpf.jvm.bytecode.Instruction;
import gov.nasa.jpf.search.Search;

public class HBSearchListener extends PropertyListenerAdapter {//extends ListenerAdapter {
   
	//static MyMap<Integer, MyMap<Key, MySet<Key>>> h_future = new MyMap<Integer, MyMap<Key, MySet<Key>>>(); // stores the h_future 
	 static ArrayList<FutureAccess> future_access = new ArrayList<FutureAccess>();
	//static FutureAccess current_future_access = null;
	 static int current_advance = -1;
	 static  MySet<Key> thread_set = new MySet<Key>();
	 static int extraHBstates=0;
		 
	/**************** helper methods *************/
  	public void instructionExecuted (JVM vm){
	  
	//      Instruction insn = vm.getLastInstruction();
	//      if ( insn != null ) Log.logInstruction("\t\t[INSTRUCTION]"+insn+":"+insn.getMnemonic()+" at "
	//   		  +insn.getLineNumber()+":"+insn.getSourceLine());    		  
	}
  /**************** SearchListener notifications *****/

  public void stateStored (Search search)
  {
	  Log.logSearch( "HBSearchListener:ASTORE ", search);
  }
  
  public void stateRestored(Search search) {
	  Log.logSearch( "HBSearchListener:RESTORE ", search);
	  current_advance = search.getStateNumber();
	 future_access.get(current_advance).setVisited();
  }

  public void stateBacktracked (Search search) {
	  Log.logSearch( "HBSearchListener:BACK ", search);
	  current_advance = search.getStateNumber();
  }
  
  public void searchStarted (Search search) {
    Log.logSearch("HBSearchListener: SEARCH STARTED ", search);
  }

  public void searchFinished (Search search) {
    Log.logSearch("HBSearchListener: SEARCH FINISHED ", search);
  }
  
  public void stateProcessed (Search search)
  {
	  Log.logSearch( "HBSearchListener:PROCESSED ", search);
	  int id = search.getStateNumber();
	  if ( id < 0 || id >= future_access.size() ) return;
	  FutureAccess fc = future_access.get(search.getStateNumber());
	  fc.setVisited(); fc.setVisited(); // mark visited > 1	
  }

  public void stateAdvanced (Search search) {
	  Log.logSearch( "HBSearchListener:ADVANCE " , search);
	  
	  int new_advance = search.getStateNumber();
  
	  if ( new_advance < 0 ) return;
	  
	  FutureAccess new_future_access=null;
	  
	  if ( future_access.size() == new_advance )
	  {
		  new_future_access = new FutureAccess(new_advance);  
		  future_access.add(new_future_access);
	  }
	  else if ( future_access.size() > new_advance )
	  {
		  new_future_access = future_access.get(new_advance);
	  }
	  else
		  assert(false):"[ERROR]invalid future_access size.";
	  
	  if ( current_advance >= 0 ) 
	  {
		  future_access.get(current_advance).addChild(new_future_access);
	  }
	  current_advance = new_advance;
	  
	  if ( search.isEndState() )
	  {
		  new_future_access.addHfuture(new MyMap<Key, MySet<Key>>());
		  new_future_access.setProcessed();
	  }
	  else if ( new_future_access.getHfuture().size() == 0 )
	  {
		  new_future_access.addHfuture(computeHfuture(current_advance));
	  }		  
  }
  
  /**************** HBsearch related interfaces *****/
  
  public static void addFutureAccessKey(Key key, boolean isThread)
  {
	  
	  if ( current_advance >= 0 ) 
		  future_access.get(current_advance).addAccess(key);
	  
	  if ( isThread )
		  thread_set.add(key);
  }
  
  /*
   * calculate h_future for given "id" state with current h
   * should return size 0 MyMap instead of null for empty set
   */
  private static MyMap<Key, MySet<Key>> computeHfuture(int id)
  {
	  MyMap<Key, MySet<Key>> h = new MyMap<Key, MySet<Key>>();
	  MySet<Key> future = future_access.get(id).getFutureAccess(null);

	  MySet<Key> staticObj = ListenerHBmanagement.getStaticObj();
	  MySet<Key> thread = new MySet<Key>();
	  for ( Key k : thread_set )
		  if ( future.contains(k) ) thread.add(k);
	  
	  for ( Key k : future )
	  {
		  if ( ListenerHBmanagement.IS_ADDR(k) )
		  {
			  MySet<Key> set = ListenerHBmanagement.CANDIDATE_AGENTS(k);
			  
			  MySet<Key> copy = new MySet<Key>(set);
			  for ( Key v : set )
			  {
				  if ( !future.contains(v)) // when the agent is not future accessed, we don't need to include it 
					  copy.remove(v);  
			  }
			  
			  if ( staticObj.contains(k) )
				  copy.addAll(thread);

			  h.put(k, copy);
		  }
	  }
	  return h;
  }
  
  public static boolean isFirstVisit(int id)
  { 
	  return id >= future_access.size() || id < 0 || future_access.get(id)==null || !future_access.get(id).isVisited();
  }
  
  /*
   * compares stored h_future with current h
   * calculate h_future for given "id" state with current h
   */
  public static boolean isCoveredState(int id, boolean isVisitMore)
  { 
	  if ( id >= future_access.size() || id < 0 ) return false;
	FutureAccess fa = future_access.get(id);
	if (  fa == null ) return false;
	
	MyMap<Key, MySet<Key>> current_h_future = computeHfuture(id);
	
	if ( fa.isCovered(current_h_future) )
	{
		if ( isVisitMore ) Log.logSearch("\t\t[NOTICE] comparison of "+id+" with stored h_future returns this is already covered.", null);
		return true;
	}
	else
	{
		if ( isVisitMore )  Log.logSearch("\t\t[NOTICE] comparison of "+id+" with stored h_future returns this should be visited again.", null);
		if ( isVisitMore ) 
		{
			fa.addHfuture(current_h_future);
			++extraHBstates;
		}
		return false;
	}
  } 
  
  public static int getAdditionalStates()
  {
	return extraHBstates;  
  }
}
