package jrf.search;

import gov.nasa.jpf.Config;
import gov.nasa.jpf.jvm.DynamicArea;
import gov.nasa.jpf.jvm.ElementInfo;
import gov.nasa.jpf.jvm.FieldInfo;
import gov.nasa.jpf.jvm.JVM;
import gov.nasa.jpf.jvm.ThreadInfo;
import gov.nasa.jpf.jvm.bytecode.FieldInstruction;
import gov.nasa.jpf.jvm.bytecode.GETFIELD;
import gov.nasa.jpf.jvm.bytecode.GETSTATIC;
import gov.nasa.jpf.jvm.bytecode.Instruction;
import gov.nasa.jpf.jvm.bytecode.LockInstruction;
import gov.nasa.jpf.jvm.bytecode.MONITORENTER;
import gov.nasa.jpf.jvm.bytecode.MONITOREXIT;
import gov.nasa.jpf.jvm.bytecode.PUTFIELD;
import gov.nasa.jpf.jvm.bytecode.PUTSTATIC;
import gov.nasa.jpf.search.heuristic.SimplePriorityHeuristic;
import jrf.hbset.util.Log;

public class HBHeuristicWithoutJRF extends SimplePriorityHeuristic  {
	static final int MAX_HEURISTIC					=Integer.MAX_VALUE;
	static final int HIGHEST						= 10;					
	static final int WRITE_FIRST_WRITTEN_BY_OTHER 	= 9;
	static final int WRITE_FIRST_WRITTEN_BY_SELF 	= 8;
	static final int READ_WRITTEN_BY_OTHER 			= 7;
	static final int READ_WRITTEN_BY_SELF 			= 4;
	static final int ACQUIRE_FIRST_NO_PRIOR_RELEASE = 6;
	static final int BASE							= 5;
	static final int ACQUIRE_WITH_PRIOR_RELEASE 	= 3;
	static final int RELEASE 						= 1;
	//static final int LOWEST							= 0;
	
	
	private final boolean WRITE_FIRST;
	private final boolean WATCH_WRITTEN;
	private final boolean AVOID_RELEASE_ACQUIRE;
	private final boolean ACQUIRE_FIRST;
	//private final boolean ADAPTIVE_DFS;
	//private final boolean ADAPTIVE_HB;
	//private final boolean HB_DFS;
	
	private JVM vm;
		  
	  public HBHeuristicWithoutJRF (Config config, JVM vm)  {
		    super(config,vm);
		    this.vm = vm;
		    String option = config.getProperty("search.heuristic.HBHeuristic");
		    
		    WRITE_FIRST=option.contains("write_first")?true:false;
			WATCH_WRITTEN=option.contains("watch_written")?true:false;
			AVOID_RELEASE_ACQUIRE=option.contains("avoid_release_acquire")?true:false;
			ACQUIRE_FIRST=option.contains("acquire_first")?true:false;
			
			Log.logSearch("@@ HBHeuristic OPTION: "+"WRITE_FIRST ="+WRITE_FIRST
						+", WATCH_WRITTEN ="+WATCH_WRITTEN
						+", AVOID_RELEASE_ACQUIRE ="+AVOID_RELEASE_ACQUIRE
						+", ACQUIRE_FIRST ="+ACQUIRE_FIRST, null);
	  }
	  
	  protected int computeHeuristicValue ()
	 {  
		 int depth = vm.getPathLength();
		 int hvalue=BASE;
		
		 ThreadInfo ti = vm.getCurrentThread();
		 //ThreadInfo to = vm.getCurrentTransition().getThreadIndex();
		 
		 String thread = ti.getClassInfo().getName()+"@"+ti.getThreadObjectRef();
		 Instruction inst = ti.getPC();
		 
		 if ( inst instanceof GETSTATIC )
		 {
			 FieldInfo fi = ((FieldInstruction)inst).getFieldInfo();
			 String cname = ((FieldInstruction)inst).getClassName();
			 ElementInfo ei = vm.getKernelState().sa.get(cname);
			 

		     int fn = fi.getFieldIndex()+1;
			 assert(fn!=0) :"Cannot decide Field Number of \""+fi.getName()+"\" from "+ei;
				
			 if ( fi.isVolatile() ) // acquire
			 {
				// if ( ListenerHBmanagement.PRIOR_RELEASE(ei.toString(), fn))
					hvalue = ACQUIRE_WITH_PRIOR_RELEASE;
				// else
				//	 hvalue = ACQUIRE_FIRST_NO_PRIOR_RELEASE;
			 }
			 else	// read
			 {
				 //if ( ListenerHBmanagement.WRITTEN_BY(thread, ei.toString(), fn))
				//	 hvalue = READ_WRITTEN_BY_SELF;
				// else
					 hvalue = READ_WRITTEN_BY_OTHER;
			 }
		 }
		 else if ( inst instanceof PUTSTATIC )
		 {
			 FieldInfo fi = ((FieldInstruction)inst).getFieldInfo();
			 String cname = ((FieldInstruction)inst).getClassName();
			 ElementInfo ei = vm.getKernelState().sa.get(cname);
			 int fn = fi.getFieldIndex()+1;
			 assert(fn!=0) :"Cannot decide Field Number of \""+fi.getName()+"\" from "+ei;
		       
			 if ( fi.isVolatile() ) // release
			 {
				 hvalue = RELEASE;
			 }
			 else// write
			 {
				 //if ( ListenerHBmanagement.WRITTEN_BY(thread, ei.toString(), fn))//, thread, ei.toString()+"."+fname) )
				//	 hvalue = WRITE_FIRST_WRITTEN_BY_SELF;
				// else
					 hvalue = WRITE_FIRST_WRITTEN_BY_OTHER;
			 }   
			
		 }
		 else if ( inst instanceof GETFIELD )
		 {
			 FieldInfo fi = ((FieldInstruction)inst).getFieldInfo();
			 ElementInfo ei = DynamicArea.getHeap().get(ti.peek());
			 int fn = fi.getFieldIndex()+1;
			 assert(fn!=0) :"Cannot decide Field Number of \""+fi.getName()+"\" from "+ei;
		       
			 if ( fi.isVolatile() ) // acquire
			 {
				// if ( ListenerHBmanagement.PRIOR_RELEASE(ei.toString(), fn))//, ei.toString()+"."+fname))
					 hvalue = ACQUIRE_WITH_PRIOR_RELEASE;
				// else
				//	 hvalue = ACQUIRE_FIRST_NO_PRIOR_RELEASE;
			 }
			 else	// read
			 {
				 //if ( ListenerHBmanagement.WRITTEN_BY(thread, ei.toString(), fn))//, thread, ei.toString()+"."+fname) )
				//	 hvalue = READ_WRITTEN_BY_SELF;
				// else
					 hvalue = READ_WRITTEN_BY_OTHER;
			 }
		 }
		 else if ( inst instanceof PUTFIELD )
		 {
			 FieldInfo fi = ((FieldInstruction)inst).getFieldInfo();
			 int storageSize = fi.getStorageSize();
			 int objRef = ti.peek( (storageSize == 1) ? 1 : 2);
			 ElementInfo ei = DynamicArea.getHeap().get(objRef);
			 int fn = fi.getFieldIndex()+1;
			 assert(fn!=0) :"Cannot decide Field Number of \""+fi.getName()+"\" from "+ei;
		       
		       if ( fi.isVolatile() ) // release
				 {
					 hvalue = RELEASE;
				 }
				 else// write
				 {
					// if ( ListenerHBmanagement.WRITTEN_BY(thread, ei.toString(), fn))//, thread, ei.toString()+"."+fname) )
					//	 hvalue = WRITE_FIRST_WRITTEN_BY_SELF;
					// else
						 hvalue = WRITE_FIRST_WRITTEN_BY_OTHER;
				 }   
		 }
		 else if ( inst instanceof MONITORENTER ) // acquire
		 {
			 ElementInfo ei = ((LockInstruction)inst).getLastLockElementInfo();
			// if ( ListenerHBmanagement.PRIOR_RELEASE(ei.toString(), 0))//, ei.toString()))
				 hvalue = ACQUIRE_WITH_PRIOR_RELEASE;
			// else
			//	 hvalue = ACQUIRE_FIRST_NO_PRIOR_RELEASE;
		 }
		 else if ( inst instanceof MONITOREXIT ) // release
		 {
			 hvalue = RELEASE;
		 }
		 
		 {
			 String hmsg;
			 switch ( hvalue )
			 {
			 case WRITE_FIRST_WRITTEN_BY_OTHER :
				 hmsg = " WRITE_FIRST_WRITTEN_BY_OTHER";
				 if ( !WRITE_FIRST ) 
					 if ( !WATCH_WRITTEN )
						 hvalue = BASE ;
					// else
					//	 hvalue = READ_WRITTEN_BY_OTHER;
				 break;
			 case READ_WRITTEN_BY_OTHER:
				 hmsg = " READ_WRITTEN_BY_OTHER";
				 if ( !WRITE_FIRST )
					 if ( !WATCH_WRITTEN )
						 hvalue = BASE ;
					 else
						 hvalue = WRITE_FIRST_WRITTEN_BY_OTHER;
				 break;
			 case WRITE_FIRST_WRITTEN_BY_SELF:
				 hmsg = " WRITE_FIRST_WRITTEN_BY_SELF";
				 if ( !WATCH_WRITTEN )
					 if ( !WRITE_FIRST )
						 hvalue = BASE ;
					 else
						 hvalue = WRITE_FIRST_WRITTEN_BY_OTHER;
				 break;
			 case READ_WRITTEN_BY_SELF:
				 hmsg = " READ_WRITTEN_BY_SELF";
				 if ( !WATCH_WRITTEN )
					 if ( !WRITE_FIRST )
						 hvalue = BASE ;
					 else
						 hvalue = READ_WRITTEN_BY_OTHER;
				 else
					 if ( !WRITE_FIRST )
						 hvalue = WRITE_FIRST_WRITTEN_BY_SELF;
				 break;
				 
			 case ACQUIRE_WITH_PRIOR_RELEASE:
				 hmsg = " ACQUIRE_WITH_PRIOR_RELEASE";
				 if ( !ACQUIRE_FIRST && !AVOID_RELEASE_ACQUIRE ) 
						 hvalue = BASE;
				 break;	 
			 case ACQUIRE_FIRST_NO_PRIOR_RELEASE:
				 hmsg = " ACQUIRE_FIRST_NO_PRIOR_RELEASE";
				 if ( !AVOID_RELEASE_ACQUIRE )
					 if ( !ACQUIRE_FIRST ) 
						 hvalue = BASE;
					 else
						 hvalue = ACQUIRE_WITH_PRIOR_RELEASE;
				 break; 
			
			 case RELEASE:
				 hmsg = " RELEASE";
				 if ( !ACQUIRE_FIRST )
					 hvalue = BASE;
				 break;
			 default :
				 hmsg = " OTHER";
			 }
			 Log.logSearch("\t@@ HeuristicValue("+thread+" at "+inst+")="+hmsg, null);
		 }
		 return MAX_HEURISTIC-((depth+1)*(HIGHEST+1)+hvalue);
	 } 

	  public void processParent ()
	  {
		  Log.logSearch("\t@@ processParent @@", null);
	  }
}
