package jrf.hbset.util;

import java.util.ArrayList;

import gov.nasa.jpf.jvm.ClassInfo;
import gov.nasa.jpf.jvm.ElementInfo;
import gov.nasa.jpf.jvm.JVM;



public class ACHSet {

	public static final boolean choice=JVM.getVM().getConfig().getBoolean("jrf.acquiring_history");
	static class ACHelement {
		public final Key threadKey;
		public final Key agentKey;
	
		ACHelement(Key thread, Key agent)
		{
			threadKey = thread;
			agentKey = agent;
		}
		
		@Override
		public boolean equals(Object o)
		{
			if ( o == null ) return false;
			if ( !(o instanceof ACHelement) ) return false;
			return threadKey.equals(((ACHelement)o).threadKey) 
					&& agentKey.equals(((ACHelement)o).agentKey)
					;
		}
		
		@Override
		public int hashCode()
		{
			return (threadKey.hashCode())<<16|agentKey.hashCode();
		}
		
		@Override
		public String toString()
		{
			//return "("+threadKey+","+agentKey+","+accessType+")";
			return "("+threadKey+","+agentKey+")";
		}
	}
	
	static final MyMap<Key, ArrayList<ACHelement>> ach = new MyMap<Key, ArrayList<ACHelement>>();
	
	public static void add(Key memory, Key thread, Key agent)//, ACCESS_TYPE a)
	{
		if ( !choice ) 
			return;
		
		//Log.logACH("trying to add ACH["+memory+"] with ("+thread+","+agent+","+a+")");
		ACHelement e = new ACHelement(thread, agent);//, a);
		if ( ach.containsKey(memory) )
		{
			ArrayList<ACHelement> list = ach.get(memory);
			
			if ( !list.contains(e) )
			{
				list.add(e);
				ach.put(memory, list);
				Log.logACH("actually added ACH["+memory+"] with "+e);
			}
		}
		else
		{
			ArrayList<ACHelement> list = new ArrayList<ACHelement>();
			list.add(e);
			ach.put(memory, list);
			Log.logACH("initially added ACH["+memory+"] with "+e);
		}
	}
	
	public static ArrayList<String> advice(Key memory, Key thread, MySet<Key> agents)
	{ 
		ArrayList<String> message = new ArrayList<String>();
		ArrayList<ACHelement> list = ach.get(memory);
		
		Log.logACH("ach["+memory+"] entry has "+(list==null?0:list.size())+" items.");
		if ( list == null ) return message;
		
		for ( ACHelement e : list )
		{
			if ( !agents.contains(e.agentKey) ) continue;
			if ( e.agentKey.field == 0 )
			{
				int index = e.agentKey.toNameString().indexOf("@");
				String clsName = e.agentKey.toNameString().substring(0, index);
				ClassInfo ci = ClassInfo.getResolvedClassInfo(clsName);

				if ( ci != null && ci.isInstanceOf("java.lang.Thread") )
					message.add("join on ("+e.agentKey.toNameStringWithLocation()+
							") before accessing ("+memory.toNameString()+
							")\n\tas ("+e.threadKey.toNameStringWithLocation()+") did.\n");
				else
					message.add("synchronize on ("+e.agentKey.toNameStringWithLocation()+
							") before accessing ("+memory.toNameString()+
							")\n\tas ("+e.threadKey.toNameStringWithLocation()+") did.\n");
			}
			else
			{
				Log.logACH("e.agentKey="+e.agentKey+"("
						+e.agentKey.toString().startsWith("java.util.concurrent.locks.ReentrantLock$NonfairSync@")
						+","
						+e.agentKey.toString().endsWith(".4")
						+")");
				if ( (e.agentKey.toString().startsWith("java.util.concurrent.locks.ReentrantLock$NonfairSync@") 
						|| e.agentKey.toString().startsWith("java.util.concurrent.locks.ReentrantLock$FairSync@") )
						&& e.agentKey.toString().endsWith(".4") )
				{
					ClassInfo ci = ClassInfo.getResolvedClassInfo("java.util.concurrent.locks.ReentrantLock");
					int index = JVM.getVM().getDynamicArea().getNext(ci, 0);
					boolean flag=false;
					while ( index != -1 )
					{
						ElementInfo ei = JVM.getVM().getDynamicArea().get(index);
						ElementInfo sync = ei.getObjectField("sync");
						
						if ( e.agentKey.toString().equals(sync.toString()+".4") )
						{
							message.add("lock ("+(new Key(ei.toString(),0)).toNameStringWithLocation()+
									") before accessing ("+memory.toNameString()+
									")\n\tas ("+e.threadKey.toNameStringWithLocation()+") did.\n");
							flag = true;
							break;
						}
						index = JVM.getVM().getDynamicArea().getNext(ci, index+1);
					}
					if ( !flag )
						message.add("read volatile variable ("+e.agentKey.toNameStringWithLocation()+
								") before accessing ("+memory.toNameString()+
								")\n\tas ("+e.threadKey.toNameStringWithLocation()+") did.\n");
				}
				else
					message.add("read volatile variable ("+e.agentKey.toNameStringWithLocation()+
									") before accessing ("+memory.toNameString()+
									")\n\tas ("+e.threadKey.toNameStringWithLocation()+") did.\n");
		
			}
		}
		return message;
	}
}
