package jrf.hbset.util;

import gov.nasa.jpf.jvm.bytecode.Instruction;

public class HBDataRaceException extends RuntimeException {
	private static Instruction currentInstruction;
	
	public static void setSourceLocation(Instruction ins)
	{
		currentInstruction = ins;
	}
	
	public static Instruction getSourceInstruction()
	{
		return currentInstruction;
	}
	
	public static String getSourceLocation()
	{
		if ( currentInstruction==null )
			return "INITIALIZER";
		else
			/*
			return currentInstruction.getFileLocation().trim()+":"+
			currentInstruction.getSourceLine().trim().split(currentInstruction.getFileLocation().trim())[0]+
			currentInstruction.getMethodInfo().getName();
			*/
			return "\""+currentInstruction.getSourceLine().trim()+
			    "\" at "+currentInstruction.getFileLocation().trim()+
			    " in ("+currentInstruction.getMethodInfo().getName()+")";
	}
	
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	Key thread;
	Key object;
	Instruction instruction;
	
	public HBDataRaceException(Key t, Key o, Instruction i)
	{
		//super("[RACE] in {"+currentInstruction+" at " +currentSourceLocation+"}");
		this.thread = t;
		this.object = o;
		this.instruction = i;
	}
	
	public Key getMemoryObject()
	{
		return object;
	}
	
	public Key getThreadObject()
	{
		return thread;
	}
	
	public String getSourceLine()
	{
		return instruction.getFileLocation().trim();
	}
	
	@Override
	public String toString()
	{
		return super.toString()+     "\n\t at THREAD      ("+thread.toNameStringWithLocation()+") "+
									 "\n\t to MEMORY      ("+object.toNameStringWithLocation()+") "+
									 "\n\t in INSTRUCTION ("+instruction+")"+
									 "\n\t of SOURCE      (\""+instruction.getSourceLine().trim()+
									 				    "\" at "+instruction.getFileLocation().trim()+")";
	}
	
	@Override
	public boolean equals(Object o)
	{
		if ( !(o instanceof HBDataRaceException) )
			return false;
		HBDataRaceException other = (HBDataRaceException)o;
		if ( other.object.equals(object) 
				&& other.thread.equals(thread) 
				&& other.instruction.toString().equals(instruction.toString()) )
			return true;
		return false;
	}
	
	public static void handle(Key t, Key o)
	{
		
		if ( Configuration.isBenign(currentInstruction.getFileLocation().trim()) )
		{
			System.out.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
			System.out.println("[BENIGN RACE] ");
			System.out.println("\t\t at THREAD("+t+")\n\t\t to MEMORY("+o+")");
			System.out.println("\t\t in INSTRUCTION("+currentInstruction+
					")\n\t\t of SOURCE "+currentInstruction.getSourceLine().trim());
			System.out.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
		}
		else
		{
			throw new HBDataRaceException(t, o, currentInstruction);
		}
	}
	
}
