package jrf.hbset.util;

import java.lang.Iterable;
import java.util.Iterator;

public class MySet<T> implements Iterable<T> { 
	
	static final int CAPACITY = 3;
	Node<T>[] element;
	
	
	public MySet()
	{
		this(CAPACITY);
	}
	
	public MySet(int capacity)
	{
		element = new Node[capacity];
	}
	
	public MySet(MySet<T> other)
	{
		element = new Node[other.size()];
		for ( T t : other )
			add(t);
	}
	
	public int size()
	{
		for (int i=0 ; i < element.length ; ++i)
			if ( element[i] == null ) return (i);
		return element.length;
	}
	
	public MySet<T> clone()
	{
		MySet<T> clone = new MySet<T>(element.length);
		
		for (Node<T> n : element)
		{
			if ( n != null )
				clone.add(n.value);
		}
		
		return clone;
	}
	
	public boolean contains(T value) {
		for (int i=0 ; i < element.length ; ++i)
		{
			if ( element[i] == null )
				return false;
				
			else if ( element[i].value.equals(value) )
				return true;
		}
		return false;
	}
	
	public boolean remove(T value) {
		int p=-1, end=element.length;
		for (int i=0 ; i < element.length ; ++i)
		{
			if ( element[i] == null )
			{
				end = i;
				break;
			}
			else if ( element[i].value.equals(value) )
				p = i;
		}
		if ( p != -1 )
		{
			element[p].value = element[end-1].value;
			element[end-1] = null;
			return true;
		}
		return false;
    }
	
	public boolean add(T value) {

		for (int i=0 ; i < element.length ; ++i)
		{
			if ( element[i] == null )
			{
				element[i] = new Node<T>(value);
				return true;
			}
			else if ( element[i].value.equals(value) )
				return false;
		}
	
		Node<T>[] expand = new Node[element.length+CAPACITY];
		for ( int i=0 ; i < element.length ; ++i)
			expand[i] = element[i];
		expand[element.length] = new Node<T>(value);
		element = expand;
		return true;
	}
	
	public boolean addAll(MySet<T> values) {
		boolean flag = false;
		for( T t : values )
			flag |= add(t);
		return flag;
	}
	
	class Node<K> {
		 K value;
		
		Node(K value)
		{
			this.value = value;
		}
	}

	public boolean containsAll(MySet<T> other) {
		for ( T t : other )
			if ( !contains(t) ) return false;
		return true;
	}

	public void clear() {
		element = new Node[CAPACITY];
	}
	
	@Override
	public String toString()
	{
		StringBuffer buf = new StringBuffer();
		String seperator = " ";
		buf.append("{");
		for( int i=0 ; i < element.length ; ++i)
			if ( element[i] != null )
			{
				buf.append(seperator+element[i]);
				seperator = ", ";
			}
		buf.append("}");
		return buf.toString();
	}
	
	@Override
	public Iterator<T> iterator() {
		return new MyIterator();
	}
	
	class MyIterator<T> implements Iterator<T> {
		Object[] nodes;
		int current=0;
		
		MyIterator()
		{
			nodes = new Object[size()];
			for ( int i=0, j=0 ; i < element.length ; ++i )
				if ( element[i] != null )
					nodes[j++] = element[i].value;
				else
					break;
			current = nodes.length;
		}
		
		@Override
		public boolean hasNext() {
			return current>0;
		}

		@Override
		public T next() {
			return (T)nodes[--current];
		}

		@Override
		public void remove() {
			nodes[current] = null;
		}
		
	}
}