package jrf.hbset.util;

import gov.nasa.jpf.jvm.FieldInfo;
import gov.nasa.jpf.jvm.ThreadInfo;

import java.lang.reflect.Modifier;


public class Utility {
	public static enum Comparison {
		NONE, SMALLER, EQUAL, BIGGER, 
	}
	
	public static Comparison compare(MySet<Key> a, MySet<Key> b)
	{
		if ( a.containsAll(b) )
		{
			if ( b.containsAll(a) )
				return Comparison.EQUAL;
			else
			{
//Set<Key> c = new java.util.HashSet<Key>(a);
//c.removeAll(b);
//System.err.println("## compare returns bigger since new has more \""+c+"\".");
				return Comparison.BIGGER;
			}
		}
		else
		{
			if ( b.containsAll(a) )
			{
//Set<Key> c = new java.util.HashSet<Key>(b);
//c.removeAll(a);
//System.err.println("## compare returns smaller since old has more \""+c+"\".");
				return Comparison.SMALLER;
			}
				
			else
				return Comparison.NONE;
		}
	}
	
	public static boolean isPrimitiveArray(String obj)
	{
		return obj.startsWith("[")&&!obj.startsWith("[L");
	}
	
	public static boolean isArray(String obj)
	{
		return obj.startsWith("[");
	}
	
	public static boolean isProtected(int modifier)
	{
		return (modifier & Modifier.PROTECTED) != 0;
	}
	
	public static boolean isAccessible(FieldInfo fi)
	{
		return fi==null || fi.isPublic() || isProtected(fi.getModifiers());
	}
	
	
	public static boolean isImmutable(boolean eiImmutable, FieldInfo fi, int fn)
	{
		if ( fi == null && fn == 0 )
			return eiImmutable;
		else if ( fi == null )
			return false;
		else
			return false;
//			return fi.isFinal();
	}
	
	
	public static String getMethodName(String obj, int fn)
	{
		Key k = new Key(obj, fn);
		String path = k.getLocation();
		if ( path == null ) path = HBDataRaceException.getSourceLocation();
		int i= path.indexOf(" in ");
		return path.substring(i+4).trim();
	}
	
	public static boolean isExcluded(String obj, int fn, FieldInfo fi)
	{
/*		System.out.println("\t\t[isExcluded] isPrimitiveArray("+obj+")="+isPrimitiveArray(obj)+ " " +
				"getMethodName("+obj+","+fn+")="+getMethodName(obj, fn)+" " +
				" equals(" +
				"\"<clinit>\"="+getMethodName(obj, fn).equals("<clinit>") +" " +
				"isInteresting="+Configuration.isInteresting(obj) 
				);
*/	
		
		// exclude java.lang.String class Object
		if ( obj.startsWith("java.lang.String@") && fn==0 )
			return true;
		else if ( obj.startsWith("java.lang.Class@") && fn==0 )
			return true;
		else if ( isPrimitiveArray(obj) && getMethodName(obj, fn).equals("(<clinit>)") && !Configuration.isInteresting(obj) )
			return true;

		
		String path = (new Key(obj, fn)).getLocation();
		
		if ( path != null && path.lastIndexOf("\" at ")!= -1 )
		{
			path = path.substring(path.lastIndexOf("\" at ")+5);
		}
		
		if ( path == null )
			return false;
		else if ( path.startsWith("java/lang/Integer.java") && getMethodName(obj, fn).equals("(<clinit>)") )
		{
			if ( (obj.startsWith("java.lang.Integer@") && fn==0) || obj.startsWith("[Ljava.lang.Integer;@") )
				return true;
			else 
				return false;
		}
		else if ( obj.equals("java.lang.Character") 
				|| obj.equals("java.lang.Integer") 
				|| obj.equals("java.lang.Short") 
				|| obj.equals("java.lang.Float") 
				|| obj.equals("java.lang.Long") 
				|| obj.equals("java.lang.Boolean") 
				|| obj.equals("java.lang.Double")
				|| obj.equals("java.lang.System")  )
		{
			if ( path.equals("INITIALIZER") && fi != null && fi.isFinal() && fi.isStatic() )
				return true;
			else 
				return false;
		}
		else
			return false;
	}
	
	// get thread object name
	// which is "name of the class"+"@"+"object number of that class"
	public static String getThreadObject(ThreadInfo ti)
	{
		return ti.getClassInfo().getName()+"@"+ti.getThreadObjectRef();
	}

}
