
package jrf.listener;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.List;
import java.util.concurrent.TimeUnit;

import jrf.hbset.util.Log;
import jrf.listener.JRFListener.JRFresult;
import jrf.search.HBBFSearch;
import jrf.search.HBDFSearch;
import jrf.search.HBHeuristicSearch;

import gov.nasa.jpf.Config;
import gov.nasa.jpf.Error;
import gov.nasa.jpf.JPF;
import gov.nasa.jpf.ListenerAdapter;
import gov.nasa.jpf.PropertyListenerAdapter;
import gov.nasa.jpf.jvm.JVM;
import gov.nasa.jpf.jvm.StateSet;
import gov.nasa.jpf.jvm.Transition;
import gov.nasa.jpf.jvm.bytecode.Instruction;
import gov.nasa.jpf.jvm.bytecode.InvokeInstruction;
import gov.nasa.jpf.report.Publisher;
import gov.nasa.jpf.report.Statistics;
import gov.nasa.jpf.search.Search;

public class PrintSearchListener extends PropertyListenerAdapter {//extends ListenerAdapter {
   
  /*** Property ****************************************************/
    
    static HashMap<String,Integer> finishingState;
    
    static {
    	finishingState = new HashMap<String, Integer>();
    	finishingState.put("herlihy_shavit.barrier.JPFDisBarrierTest", 84);
    	finishingState.put("herlihy_shavit.barrier.JPFStaticTreeBarrierTest", 102);
    	finishingState.put("herlihy_shavit.hash.JPFCoarseHashSetTest" , 109);
    	finishingState.put("herlihy_shavit.hash.JPFLockFreeHashSetTest" , 208);
    	finishingState.put("herlihy_shavit.hash.JPFRefinableHashSetTest" , 177);
    	finishingState.put("herlihy_shavit.hash.JPFStripedHashSetTest" , 91);
    	finishingState.put("herlihy_shavit.hash.JPFTCuckooHashSetTest",936 );
    	finishingState.put("herlihy_shavit.lists.JPFLazyListTest" , 136);
    	finishingState.put("herlihy_shavit.lists.JPFOptimisticListTest" , 132);
    	finishingState.put("herlihy_shavit.mutex.JPFBakeryTest" , 87);
    	finishingState.put("herlihy_shavit.mutex.JPFFilterTest" , 61);
    	finishingState.put(	"herlihy_shavit.mutex.JPFLockFreeQueueTest" , 43);
    	finishingState.put(	"herlihy_shavit.mutex.JPFPetersonTest" , 61);
    	finishingState.put(	"herlihy_shavit.spin.JPFALockTest" , 48);
    	finishingState.put("herlihy_shavit.spin.JPFCLHLockTest" , 40);
    	finishingState.put("herlihy_shavit.spin.JPFMCSLockTest" , 36);
    	finishingState.put("herlihy_shavit.spin.JPFCorrectedMCSLockTest" , 51);
    	finishingState.put(	"herlihy_shavit.steal.JPFDEQueueTest" , 23);
    	finishingState.put(	"herlihy_shavit.queue.JPFUnboundedQueueTest" , 138);
    	finishingState.put("org.amino.jpftest.IteratorTest_EBDeque_3" , 26);
    	finishingState.put("org.amino.jpftest.IteratorTest_LockFreeDeque_3" , 26);
    	finishingState.put("org.amino.jpftest.IteratorTest_LockFreeList_3" ,45 );
    	finishingState.put("org.amino.jpftest.IteratorTest_LockFreeOrderedList_3" , 45);
    	finishingState.put("org.amino.jpftest.IteratorTest_LockFreePriorityQueue_3" , 32);
    	finishingState.put("org.amino.jpftest.IteratorTest_LockFreeQueue_3" ,309 );
    	finishingState.put("org.amino.jpftest.IteratorTest_LockFreeSet_3" , 51);
    	finishingState.put("org.amino.jpftest.QueueTest7" , 54);
    	finishingState.put("google.concurrentBarrier.main.LinearSenseBarrierVolatileTest" , 112);
    	finishingState.put("google.concurrentBarrier.main.LinearSenseBarrierTest" , 115);
    	finishingState.put("google.concurrentBarrier.main.SimpleBarrierTest" , 60);
    	finishingState.put("google.concurrentBarrier.main.SenseBarrierTest" , 95);
    	finishingState.put("google.concurrentBarrier.main.SenseBarrierWithWaitTest" , 101);
    	finishingState.put("google.concurrentBarrier.main.TreeBarrierTest" , 134);
    	finishingState.put("google.concurrentBarrier.main.LockBarrierTest" , 109);
    	finishingState.put("google.concurrentBarrier.main.CyclicBarrierTest" , 76);
    	finishingState.put(	"google.concurrentBarrier.main.BinaryStaticTreeBarrierTest" , 142);
    	finishingState.put(	"google.concurrentBarrier.main.SplittedSenseBarrierTest" , 127);
    	finishingState.put(	"jgf.section1.JGFBarrierBench", 159);
    	finishingState.put(	"jgf.section1.JGFSyncBench" , 46);
    	finishingState.put(	"jgf.section2.lufact.JGFLUFactBench" , 39);
    	finishingState.put(	"jgf.section2.sor.JGFSORBench" , 86);
    	finishingState.put(	"jgf.section3.moldyn.JGFMolDynBench" , 1901);
    	finishingState.put(	"jgf.section3.montecarlo.JGFMonteCarloBench" , 178);					
    }
    
  	public boolean check (Search search, JVM vm)
	{
  		String target = JVM.getVM().getConfig().getTarget();
  		Integer limit = finishingState.get(target);
		if (  limit!= null && search.getStateNumber() >= limit )
			return false;
		else
			return true;
	}

  	 static long  startTime = System.nanoTime();
  	 
  	 private void writeResultToFile()
  	 { 
  		System.err.println("writeResultToFile!!");
  		 String resultFile = "result/RESULT";
  		 long duration = TimeUnit.SECONDS.convert((System.nanoTime()-startTime), TimeUnit.NANOSECONDS);
		 String detectionT = new String((duration/3600)+":"+((duration%3600)/60)+":"+(duration%60));
  		 try {
  			 PrintWriter pw = new PrintWriter(new FileOutputStream(resultFile, true));

  			 Config config = JVM.getVM().getConfig();
  			 Statistics stat = JVM.getVM().getJPF().getReporter().getStatistics();

  			 String msg = config.getTarget();
  			 msg += " , "+"JPF";
  			 msg += " , "+stat.newStates;
  			 msg += " , "+0;
  			 msg += " , "+stat.maxDepth;
  			 msg += " , "+detectionT+"s";
  			 msg += " , "+(stat.maxUsed >>20) + "MB";
  			 msg += " , ";
  			 msg += " , "; 
  			 System.err.println("JPFresult = "+msg+"\n\n");			 
  			 pw.println(msg);
  			 pw.flush();
  		 } catch (FileNotFoundException e) {
  			 e.printStackTrace();
  		 }
	 }
		
	public String getErrorMessage ()
	{
		String target = JVM.getVM().getConfig().getTarget();
		return "SEARCH FINISHING STATE "+finishingState.get(target)+" for "+target+" reached.";
	}
	
  /**************** SearchListener notifications *****/

  public void stateStored (Search search)
  {
	  Log.logSearch( "ASTORE ", search);
  }
  
  public void stateRestored(Search search) {
	  Log.logSearch( "RESTORE ", search);
  }

  public void stateBacktracked (Search search) {
	  Log.logSearch( "BACK ", search);
  }
  
  public void choiceGeneratorSet (JVM vm) 
  {
	  //Log.logSearch( "ChoiceGeneratorSET "+vm.getLastChoiceGenerator().toString());
  }
  public void choiceGeneratorAdvanced (JVM vm) 
  {
	  //Log.logSearch( "ChoiceGeneratorADVANCED "+vm.getLastChoiceGenerator().toString());
  }
  public void choiceGeneratorProcessed (JVM vm) 
  {
	  //Log.logSearch( "ChoiceGeneratorPROCESSED "+vm.getLastChoiceGenerator().toString());
  }
  
  public void searchStarted (Search search) {
    Log.logSearch(" SEARCH STARTED ", search);
    search.addProperty(this);
  }

  public void searchFinished (Search search) {
    Log.logSearch(" SEARCH FINISHED ", search);
    writeResultToFile();
  }
  public void stateProcessed (Search search)
  {
	  Log.logSearch( "PROCESSED ", search);
  }
  
  public void propertyViolated (Search search) {
    List<Error> errors = search.getErrors();

    for (int i=0; i<errors.size(); i++) {
      Log.logSearch("PROPERTY VIOLATED (" + (i+1)+"/"+errors.size()+")"+errors.get(i), search);
    }
  }
  
  public void searchConstraintHit(Search search) {
    //Log.logSearch("constraint hit:" + search.getSearchConstraint());
  }

  public void stateAdvanced (Search search) {
	  Log.logSearch( search.hasNextState() ? "ADVANCE " : "* ", search);
  }
   
  /**************** helper methods *************/
  	public void instructionExecuted (JVM vm){
	  
//	      Instruction insn = vm.getLastInstruction();
//	      if ( insn != null ) Log.logInstruction("\t\t[INSTRUCTION]"+insn+":"+insn.getMnemonic()+" at "
//	   		  +insn.getLineNumber()+":"+insn.getSourceLine());
	    		  
	  }
}
