package jrf.search;

import java.util.LinkedList;

import jrf.hbset.ListenerHBmanagement;

import gov.nasa.jpf.Config;
import gov.nasa.jpf.jvm.JVM;
import gov.nasa.jpf.search.heuristic.HeuristicSearch;
import gov.nasa.jpf.search.heuristic.HeuristicState;
import gov.nasa.jpf.search.heuristic.PrioritizedState;

public abstract class HBPriorityHeuristic  extends HeuristicSearch {
	int maxQueueSize;
	LinkedList<PrioritizedState> queue;
	
	public HBPriorityHeuristic (Config config, JVM vm) {
		super(config,vm);
		maxQueueSize = config.getInt("search.heuristic.queue_limit", 1024);
		if (maxQueueSize < 0){
			maxQueueSize = Integer.MAX_VALUE;
		}
		queue = new LinkedList<PrioritizedState>();    
	}

	protected abstract int computeHeuristicValue ();

	protected int computeAstarPathCost (JVM vm) {
		return vm.getPathLength();
	}

	protected HeuristicState queueCurrentState () {
		int heuristicValue;

		if (vm.isInterestingState()) {
			heuristicValue = 0;
		} else if (vm.isBoringState()) {
			heuristicValue = Integer.MAX_VALUE;
		} else {
			heuristicValue = computeHeuristicValue();

			if (useAstar) {
				// <2do> we probably don't want this for isInteresting/isBoring?
				heuristicValue += computeAstarPathCost(vm);
			}
		}
		PrioritizedState hState = new PrioritizedState(vm,heuristicValue);
		add(hState);

		return hState;
	}

	protected HeuristicState getNextQueuedState () {
		if (queue.size() == 0) {                      // the dreaded Java 1.5 version
			return null;
		}
		HeuristicState hState = queue.get(0);

		if (isBeamSearch) {
			queue.clear();
		} else {
			queue.remove(0);   
		}
		return hState;
	}

	public int getQueueSize() {
		return queue.size();
	}
	
	public boolean add(PrioritizedState s) {
		if (queue.size() < maxQueueSize) { // limit not yet reached
			int i;
			if ( ListenerHBmanagement.is_FIFO_Order() )
			{
				for(  i=queue.size() ; i > 0  ; --i)
				{
					/*
					if ( s.compareTo(queue.get(i-1)) == 0 ) 
						System.err.println("## new state("+s.getStateId()+" with "+s.getPriority()+
								") = old state("+queue.get(i-1).getStateId()+" with "+queue.get(i-1).getPriority()+")"+
								" QUEUE_SIZE="+queue.size()+", i="+i); //+":"+queue);
								*/
					if ( s.compareTo(queue.get(i-1)) >= 0 ) break; 
				}
			}
			else
			{
				for(  i=0 ; i < queue.size()  ; ++i)
					if ( s.compareTo(queue.get(i)) <= 0 ) break; 
			}
	
		//	System.out.println("HBSearchListener: ADD "+s.getStateId()+"("+s.getPriority()+") to "+i+
		//			":FIFO="+ListenerHBmanagement.is_FIFO_Order()+",next="+(queue.size()>i?s.compareTo(queue.get(i)):"")+
		//			",prev="+(i>0?s.compareTo(queue.get(i-1)):""));
			queue.add(i, s);
			return true;
		} else {
			PrioritizedState last = queue.get(queue.size()-1);
			if (s.compareTo(last) < 0) {
				queue.remove(queue.size()-1);
				return queue.add(s);
			} else {
				// don't add with this priority value
				return false;
			}
		}
	}

	public boolean isQueueLimitReached() {
		return queue.size() >= maxQueueSize;
	}
}
